/** 
 * PersonRoles.java --- Roles as subobjects 
 */

import java.util.*;

/**
 * @author Uwe Assmann
 * How different people share features (e.g., names), but drink and behave differently. 
 * Showing the semantics of the overriding of method drinkBeer() with Roles.
 * Implementation pattern: Roles with generic backlinks.
 * @date 2013-04-20
 */
public class PersonRoles {
    // The PersonWithRoleFactory creates all persons and their roles. It is
    // allocated once. 
    public static PersonWithRoleFactory factory; 
    public static PersonWithRoleFactory getPersonWithRoleFactory() { return factory; }

    // Main program. 
    public static void main(String args[]) {
	// Allocating the factory for roles.
	factory = new PersonWithRoleFactory();

	// (1) Create Prof. Aßmann with his role
	Person UweAssmann = factory.createProfessor("Uwe Aßmann");

	// See what he drinks, according to his role
	// uweAsProfessor is a role variable (subobject) of UweAssmann
	// Unfortunatly, in Java this needs casting
	Professor uweAsProfessor = (Professor)UweAssmann.asRole();
	uweAsProfessor.drinkBeer();

	// (2) Make Person Victoria Miller a Beginner of Prof. Aßmann
	Person VictoriaMiller = factory.createBeginner("Victoria Miller");
	Beginner vicciAsBeginner = (Beginner)VictoriaMiller.asRole();
	vicciAsBeginner.drinkBeer();
	vicciAsBeginner.startStudy(uweAsProfessor);
	vicciAsBeginner.study();

	// (3) mature Victoria Miller, exchange her role from Beginner to Student
	uweAsProfessor.doExam(vicciAsBeginner);
	// The old role as beginner is invalid. Get the new role as Student
	Student vicciAsStudent = (Student)VictoriaMiller.asRole();
	vicciAsStudent.drinkBeer();
    }
}

/**
 * The generic role class providing a core object link (player link)
 * This net is typed, i.e., the player is restricted to be a subclass of Person.
 */
class Role<myCore extends Person> {
    myCore player; 
    myCore getPlayer() { return player; }
    // For core attributes: backward delegation from role to core
    String getName() { return player.getName(); }
    // Now, this method is role-specific. It returns name of the core object with ":" core type, and role type
    String getTypedName() { return player.getTypedName()+":"+getClass().getName(); }
}

/** 
 * Base class Person.
 */
class Person {
    String name; // internal attribute name; not visible
    // Constuctor 
    public Role<Person> has_a_role;
    Role<Person> asRole() { return has_a_role; }
    public Person() { }
    public Person(String a_name) { setName(a_name); }
    // Attribute query: read the name of a Person
    String getName() { return name; }
    // This method uses the Class prototype to get the name of the class (class Class)
    String getClassName() { return getClass().getName(); }
    // This method returns name of the object with ":" and type
    String getTypedName() { return getName()+":"+getClass().getName(); }
    // Attribute modifier: change the name of a Person
    void setName(String a_name) { name = a_name; }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Limo.... Emptying Limo...");
    }
}
/**
 *  Man and Woman are subclasses of Person 
 */
class Man extends Person {   }
class Woman extends Person {   }

class PersonWithRoleFactory {
    // Factory Method to create professors
    public Person createProfessor(String aName) {
	Person core = new Person();
	Professor role = new Professor(aName,core);
	core.has_a_role = role;
	return core;
    }
    public Person createBeginner(String aName) {
	Person core = new Person();
	Beginner role = new Beginner(aName,core);
 	core.has_a_role = role;
	return core;
    }
    public Person createStudent(String aName) {
	Person core = new Person();
	Student role = new Student(aName,core);
 	core.has_a_role = role;
	return core;
    }
}
class Professor extends Role < Person > {
    int numberOfStudents; 
    // Crowd of students OR alumni (polymorphic..)
    Person students[];

    // Constructors
    public Professor() { }
    public Professor(String a_name, Person core) { 
	player = core;
	player.setName(a_name);
	numberOfStudents = 0; 
	// Create array of 200 students
	students = new Person[200];
    }
    // For role methods: work on role attributes
    int getNumberOfStudents() { return numberOfStudents; }
    void setNumberOfStudents(int a_numberOfStudents) { numberOfStudents = a_numberOfStudents; }
    void incrNumberOfStudents() { numberOfStudents++; }
    void registerStudent(Beginner aBeginner) { 
	if (numberOfStudents < 200) {
	    System.out.println("Welcome, dear student "+aBeginner.getName()+" to my course ST!...");
	    students[numberOfStudents] = aBeginner.getPlayer(); 
	    incrNumberOfStudents();
	    System.out.println(getName()+": I have now "+getNumberOfStudents()+" students in my course ST...");
	} else {
	    System.out.println("Sorry, course is full...");
	}
    }
    /**
     * Important: the morphing of doExam does not need a return value any more
     * because the core object stays the same
     */ 
    public void doExam(Beginner aBeginner) { 
	System.out.println(aBeginner.getTypedName()+" is being examined...      done...");
	for (int index = 0; index <200; index++) {
	    Person curPerson = (Person)students[index];
	    if ( curPerson.getName().equals(aBeginner.getName()) ) {
		// Student found; change from Beginner to Student
// 		System.out.println("   Student "+aBeginner.getTypedName()+" under index "+index+" found");
		if (curPerson.has_a_role instanceof Beginner) {
		    // re-allocating the role, maturing it to Student
		    Student newStudent = new Student(aBeginner.getName(), curPerson);
  		    curPerson.has_a_role = newStudent;
		    System.out.println("---- "+newStudent.getTypedName()+" matured");
		    return;
		} 
		// otherwise, no action required
	    }
	}
	// Student not found
	System.out.println("ERROR: Student "+aBeginner.getTypedName()+" not found in student list");
	return;
    }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Leffe.... Emptying Leffe...");
    }
} 

/**
 * Beginner is a role class of a Person.
 */
class Beginner extends Role<Person> {
    public Beginner() { }
    public Beginner(String a_name, Person core) { 
	player = core;
	player.setName(a_name);
    }
    public void startStudy(Professor prof) { 
	prof.registerStudent(this);
	System.out.println("---- Starting to study with Prof. "+prof.getName()+" in course ST...");
    }
    public void study() { 
	System.out.println("Working hard .... in course ST...");
	System.out.println("  Writing up questions...");
	System.out.println("  Asking questions...");
	System.out.println("Working hard .... in course ST...");
    }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Clausthaler alkoholfrei.... Emptying Clausthaler alkoholfrei...");
    }
}

/**
 * Student is a role class of a Person.
 */
class Student extends Role<Person> {
    public Student() { }
    public Student(String a_name, Person core) { 
	player = core;
	player.setName(a_name);
    }
    void drinkBeer() { 
	System.out.println(getTypedName()+" opening Radeberger.... Emptying Radeberger...");
    }
}

